@@ -1,32 +1,34 @@
-# mssql/pymssql.py
-# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
+# dialects/mssql/pymssql.py
+# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
+# mypy: ignore-errors
+
 
 """
 .. dialect:: mssql+pymssql
     :name: pymssql
     :dbapi: pymssql
     :connectstring: mssql+pymssql://<username>:<password>@<freetds_name>/?charset=utf8
-    :url: http://pymssql.org/
 
 pymssql is a Python module that provides a Python DBAPI interface around
-`FreeTDS <http://www.freetds.org/>`_.  Compatible builds are available for
-Linux, MacOSX and Windows platforms.
+`FreeTDS <https://www.freetds.org/>`_.
+
+.. versionchanged:: 2.0.5
+
+    pymssql was restored to SQLAlchemy's continuous integration testing
 
-Modern versions of this driver work very well with SQL Server and
-FreeTDS from Linux and is highly recommended.
 
 """  # noqa
 import re
 
 from .base import MSDialect
 from .base import MSIdentifierPreparer
-from ... import processors
 from ... import types as sqltypes
 from ... import util
+from ...engine import processors
 
 
 class _MSNumeric_pymssql(sqltypes.Numeric):
@@ -39,14 +41,16 @@ class _MSNumeric_pymssql(sqltypes.Numeric):
 
 class MSIdentifierPreparer_pymssql(MSIdentifierPreparer):
     def __init__(self, dialect):
-        super(MSIdentifierPreparer_pymssql, self).__init__(dialect)
+        super().__init__(dialect)
         # pymssql has the very unusual behavior that it uses pyformat
         # yet does not require that percent signs be doubled
         self._double_percents = False
 
 
 class MSDialect_pymssql(MSDialect):
+    supports_statement_cache = True
     supports_native_decimal = True
+    supports_native_uuid = True
     driver = "pymssql"
 
     preparer = MSIdentifierPreparer_pymssql
@@ -57,7 +61,7 @@ class MSDialect_pymssql(MSDialect):
     )
 
     @classmethod
-    def dbapi(cls):
+    def import_dbapi(cls):
         module = __import__("pymssql")
         # pymmsql < 2.1.1 doesn't have a Binary method.  we use string
         client_ver = tuple(int(x) for x in module.__version__.split("."))
@@ -73,8 +77,8 @@ class MSDialect_pymssql(MSDialect):
         return module
 
     def _get_server_version_info(self, connection):
-        vers = connection.scalar("select @@version")
-        m = re.match(r"Microsoft .*? - (\d+).(\d+).(\d+).(\d+)", vers)
+        vers = connection.exec_driver_sql("select @@version").scalar()
+        m = re.match(r"Microsoft .*? - (\d+)\.(\d+)\.(\d+)\.(\d+)", vers)
         if m:
             return tuple(int(x) for x in m.group(1, 2, 3, 4))
         else:
@@ -86,7 +90,7 @@ class MSDialect_pymssql(MSDialect):
         port = opts.pop("port", None)
         if port and "host" in opts:
             opts["host"] = "%s:%s" % (opts["host"], port)
-        return [[], opts]
+        return ([], opts)
 
     def is_disconnect(self, e, connection, cursor):
         for msg in (
@@ -99,20 +103,24 @@ class MSDialect_pymssql(MSDialect):
             "message 20006",  # Write to the server failed
             "message 20017",  # Unexpected EOF from the server
             "message 20047",  # DBPROCESS is dead or not enabled
+            "The server failed to resume the transaction",
         ):
             if msg in str(e):
                 return True
         else:
             return False
 
-    def set_isolation_level(self, connection, level):
+    def get_isolation_level_values(self, dbapi_connection):
+        return super().get_isolation_level_values(dbapi_connection) + [
+            "AUTOCOMMIT"
+        ]
+
+    def set_isolation_level(self, dbapi_connection, level):
         if level == "AUTOCOMMIT":
-            connection.autocommit(True)
+            dbapi_connection.autocommit(True)
         else:
-            connection.autocommit(False)
-            super(MSDialect_pymssql, self).set_isolation_level(
-                connection, level
-            )
+            dbapi_connection.autocommit(False)
+            super().set_isolation_level(dbapi_connection, level)
 
 
 dialect = MSDialect_pymssql
